<?php
/**
 * WC_MNM_Store_API class
 *
 * @package  WooCommerce Mix and Match Products/REST API
 * @since    2.0.0
 * @version  2.7.9
 */

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

use Automattic\WooCommerce\StoreApi\Schemas\V1\ProductSchema;
use Automattic\WooCommerce\StoreApi\Exceptions\RouteException;
use Automattic\WooCommerce\StoreApi\Schemas\V1\CartItemSchema;
use Automattic\WooCommerce\StoreApi\Exceptions\InvalidCartException;

/**
 * Extends the store public API with container related data for each container parent and child item.
 */
class WC_MNM_Store_API {

	/**
	 * Stores the cart item key of the last child item.
	 *
	 * @var string
	 */
	private static $last_child_item_key;

	/**
	 * Plugin Identifier, unique to each plugin.
	 *
	 * @var string
	 */
	const IDENTIFIER = 'mix_and_match';

	/**
	 * Bootstraps the class and hooks required data.
	 */
	public static function init() {

		self::extend_store();

		// Aggregate cart item prices/subtotals and filter min/max/multipleof quantities.
		add_filter( 'rest_request_after_callbacks', array( __CLASS__, 'filter_cart_item_data' ), 10, 3 );
		add_filter( 'woocommerce_hydration_request_after_callbacks', array( __CLASS__, 'filter_cart_item_data' ), 10, 3 );

		// Register custom configuration data.
		add_filter( 'woocommerce_store_api_add_to_cart_data', array( __CLASS__, 'add_to_cart_data' ), 10, 2 );

		// Validate container add to cart in the Store API and add cart errors.
		add_action( 'woocommerce_store_api_validate_add_to_cart', array( __CLASS__, 'validate_add_to_cart_item' ), 10, 2 );

		// Validate container in the Store API and add cart errors.
		add_action( 'woocommerce_store_api_validate_cart_item', array( __CLASS__, 'validate_cart_item' ), 10, 2 );

		// Remove quantity selectors from child items.
		add_filter( 'woocommerce_store_api_product_quantity_editable', array( __CLASS__, 'product_quantity_editable' ), 10, 3 );

		// Prevent removal of child items.
		add_action( 'woocommerce_remove_cart_item', array( __CLASS__, 'prevent_child_item_removal' ), 10, 2 );

	}

	/**
	 * Registers the actual data into each endpoint.
	 */
	public static function extend_store() {

		if ( ! function_exists( 'woocommerce_store_api_register_endpoint_data' ) ) {
			return;
		}

		// Extend Regular Mix and Match Product Data.
		woocommerce_store_api_register_endpoint_data(
			array(
				'endpoint'        => ProductSchema::IDENTIFIER,
				'namespace'       => 'mix_and_match',
				'data_callback'   => array( __CLASS__, 'extend_product_data' ),
				'schema_callback' => array( __CLASS__, 'extend_product_schema' ),
				'schema_type'     => ARRAY_A,
			)
		);

		// Extend Cart Item Data.
		woocommerce_store_api_register_endpoint_data(
			array(
				'endpoint'        => CartItemSchema::IDENTIFIER,
				'namespace'       => self::IDENTIFIER,
				'data_callback'   => array( __CLASS__, 'extend_cart_item_data' ),
				'schema_callback' => array( __CLASS__, 'extend_cart_item_schema' ),
				'schema_type'     => ARRAY_A,
			)
		);
	}

	/*-----------------------------------------------------------------------------------*/
	/*  Mix and Match Product Data                                                       */
	/*-----------------------------------------------------------------------------------*/

	/**
	 * Register parent/child product data into cart/items endpoint.
	 * 
	 * @since 2.6.0
	 *
	 * @param WC_Product  $product
	 * @return array $item_data
	 */
	public static function extend_product_data( $product ) {

		$item_data = [];

		if ( wc_mnm_is_product_container_type( $product ) ) {
			$item_data['single_add_to_cart_text'] = $product->single_add_to_cart_text();
			$item_data['layout_override']      = $product->has_layout_override();
			$item_data['layout']               = $product->get_layout();
			$item_data['form_location']        = $product->get_add_to_cart_form_location();
			$item_data['content_source']       = $product->get_content_source();
			$item_data['child_category_ids']   = $product->get_child_category_ids();
			$item_data['child_categories']     = $item_data['content_source'] !== 'products' ? get_terms( array( 'include' => $product->get_child_category_ids() ) ) : [];
			$item_data['min_container_size']   = $product->get_min_container_size();
			$item_data['max_container_size']   = $product->get_max_container_size();
			$item_data['priced_per_product']   = $product->is_priced_per_product();
			$item_data['packing_mode']         = $product->get_packing_mode();
			$item_data['shipped_per_product']  = ! $product->is_packed_together();
			$item_data['weight_cumulative']    = $product->is_weight_cumulative();
			$item_data['discount']             = $product->get_discount();
			$item_data['child_items']          = self::prepare_child_items_response( $product );
        }

		return $item_data;
	}

	/**
	 * Register variations schema into product endpoint.
	 * 
	 * @since 2.6.0
	 *
	 * @return array Registered schema.
	 */
	public static function extend_product_schema() {

		return array(
			'single_add_to_cart_text' => array(
				'description' => esc_html__( 'Single-product details page add to cart form text. Applicable only for Mix and Match type products.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'string',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'layout_override' => array(
				'description' => esc_html__( 'Has product-specific layouts that override global setting. Applicable only for Mix and Match type products.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'boolean',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'layout'              => array(
				'description' => esc_html__( 'Single-product details page layout. Applicable only for Mix and Match type products.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'string',
				'enum'        => array_keys( WC_Product_Mix_and_Match::get_layout_options() ),
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'form_location' => array(
				'description' => esc_html__( 'Single-product details page add to cart form location. Applicable only for Mix and Match type products.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'string',
				'enum'        => array_keys( WC_Product_Mix_and_Match::get_add_to_cart_form_location_options() ),
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'content_source' => array(
				'description' => esc_html__( 'Source of child products. Applicable only for Mix and Match type products.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'string',
				'enum'        => array( 'products', 'categories' ),
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'child_category_ids' => array(
				'description' => esc_html__( 'List of child categories allowed in this product.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'array',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'child_items'     => array(
				'description' => esc_html__( 'List of child items contained in this product.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'array',
				'context'     => array( 'view' ),
				'readonly'    => true,
				'items'       => array(
					'type'       => 'object',
					'properties' => array(
						'availability' => array(
							'description' => esc_html__( 'Child item product availability.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'object',
							'properties' => array(
								'availability' => array(
									'description' => esc_html__( 'Child product available text', 'woocommerce-mix-and-match-products' ),
									'type'        => 'string',
									'context'     => array( 'view' ),
									'readonly'    => true,
								),
								'class'            => array(
									'description' => esc_html__( 'Child item product available html class.', 'woocommerce-mix-and-match-products' ),
									'type'        => 'string',
									'context'     => array( 'view' ),
									'readonly'    => true,
								),
							),
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'catalog_visibility'            => array(
							'description' => esc_html__( 'Child item product cataglog visibility.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'string',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'child_id'            => array(
							'description' => esc_html__( 'Child product|variation ID.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'child_item_id'            => array(
							'description' => esc_html__( 'Child item ID.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'max_qty' => array(
							'description' => esc_html__( 'Child item maximum quantity.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'min_qty'   => array(
							'description' => esc_html__( 'Child item minimum quantity.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'step_qty' => array(
							'description' => esc_html__( 'Child item step quantity.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'price_html'            => array(
							'description' => esc_html__( 'Child item product price html text.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'string',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'product_id'   => array(
							'description' => esc_html__( 'Child product ID.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'images'            => array(
							'description' => esc_html__( 'Child item product images.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'array',
							'context'     => array( 'view' ),
							'readonly'    => true,
							'items'       => array(
								'type'       => 'object',
								'properties'  => array(
									'alt'            => array(
										'description' => esc_html__( 'Image attachment alt text.', 'woocommerce-mix-and-match-products' ),
										'type'        => 'string',
										'context'     => array( 'view' ),
										'readonly'    => true,
									),
									'id'            => array(
										'description' => esc_html__( 'Image attachment ID.', 'woocommerce-mix-and-match-products' ),
										'type'        => 'integer',
										'context'     => array( 'view' ),
										'readonly'    => true,
									),
									'name'            => array(
										'description' => esc_html__( 'Image attachment name.', 'woocommerce-mix-and-match-products' ),
										'type'        => 'string',
										'context'     => array( 'view' ),
										'readonly'    => true,
									),
									'src'            => array(
										'description' => esc_html__( 'Image attachment src.', 'woocommerce-mix-and-match-products' ),
										'type'        => 'string',
										'context'     => array( 'view' ),
										'readonly'    => true,
									),
								)
							)
						),
						'stock_status' => array(
							'description' => esc_html__( 'Child item product stock status.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'string',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'in_stock' => array(
							'description' => esc_html__( 'Child item product is in stock.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'boolean',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'name'            => array(
							'description' => esc_html__( 'Child item product title', 'woocommerce-mix-and-match-products' ),
							'type'        => 'string',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'permalink'            => array(
							'description' => esc_html__( 'Child item product permalink.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'string',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'purchasable' => array(
							'description' => esc_html__( 'Child item product purchasble status.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'boolean',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'qty' => array(
							'description' => esc_html__( 'Child item quantity.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'short_description'            => array(
							'description' => esc_html__( 'Child item product short description.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'string',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
						'variation_id' => array(
							'description' => esc_html__( 'Child variation ID.', 'woocommerce-mix-and-match-products' ),
							'type'        => 'integer',
							'context'     => array( 'view' ),
							'readonly'    => true,
						),
					)
				)
			),
			'min_container_size'   => array(
				'description' => esc_html__( 'Minimum container size.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'integer',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'max_container_size'   => array(
				'description' => esc_html__( 'Maximum container quantity.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'mixed',
				'oneOf'       => array(
					'type' => 'integer',
					'type' => null,
				),
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'discount' => array(
				'description' => esc_html__( 'Indicates the percentage discount to apply to each child product when per-product pricing is enabled.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'string',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'priced_per_product' => array(
				'description' => esc_html__( 'Indicates whether the container price is calculated from the price of the selected child products.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'boolean',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'packing_mode' => array(
				'description' => esc_html__( 'Indicates how the child products are packed/shipped.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'boolean',
				'enum'        => array( 'virtual', 'together', 'separate', 'separate_plus' ),
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'shipped_per_product' => array(
				'description' => esc_html__( 'Deprecated: Indicates whether the child products are shipped individually.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'boolean',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
			'weight_cumulative' => array(
				'description' => esc_html__( 'Shipping weight calculation mode.', 'woocommerce-mix-and-match-products' ),
				'type'        => 'boolean',
				'context'     => array( 'view' ),
				'readonly'    => true,
			),
		);
	}

	/**
	 * Convert child items into a REST readable array.
	 * 
	 * @since 2.6.0
	 *
	 * @param WC_Product     $product  Product object.
	 * @return array
	 */
	private static function prepare_child_items_response( $product ) {
		$child_items = $product->get_child_items();
		$response_items = array();
		foreach( $child_items as $child_item ) {

			/**
			 * Individual child item Store API response.
			 *
			 * @param array $response
			 * @param  obj WC_MNM_Child_Item $child_item of child item
			 * @param  obj WC_Product_Mix_and_Match $product
			 */	
			$response_items[] = apply_filters( 'wc_mnm_child_item_store_api_response', array(
				'child_item_id'      => $child_item->get_child_item_id(),
				'child_id'           => $child_item->get_variation_id() ? $child_item->get_variation_id() : $child_item->get_product_id(),
				'product_id'         => $child_item->get_product_id(),
				'variation_id'       => $child_item->get_variation_id(),
				'min_qty'            => $child_item->get_quantity( 'min' ),
				'max_qty'            => $child_item->get_quantity( 'max' ),
				'step_qty'           => $child_item->get_quantity( 'step' ),
				'qty'                => $child_item->get_quantity(),
				'availability'       => $child_item->get_availability(),
				'purchasable'        => $child_item->get_product()->is_purchasable(),
				'in_stock'           => $child_item->get_product()->is_in_stock(),
				'stock_status'       => $child_item->get_product()->get_stock_status(),
				'price_html'         => $child_item->get_product()->get_price_html(),
				'catalog_visibility' => $child_item->get_product()->get_catalog_visibility(),
				'images'             => self::get_images( $child_item, $product ),
				'name'               => $child_item->get_product()->get_name(),
				'permalink'          => $child_item->get_product()->get_permalink(),
				'short_description'  => $child_item->get_product()->get_short_description(),
                'category_ids'       => $child_item->get_product()->get_category_ids()
			), $child_item, $product );
		}

		return $response_items;
	}


	

	/**
	 * Get the images for a child item's product.
	 * 
	 * @since 2.6.0
	 *
	 * @param WC_MNM_Child_Item.
	 * @param WC_Product_Mix_and_Match
	 *
	 * @return array
	 */
	private static function get_images( $child_item, $container_product ) {

		$product = $child_item->get_product();

		$image_size = 'woocommerce_thumbnail';

		if ( has_filter( 'wc_mnm_api_child_item_thumbnail_size' ) ) {

			wc_deprecated_function( 'wc_mnm_api_child_item_thumbnail_size', '2.7.0', 'wc_mnm_api_child_item_thumbnail_sizes, nb: Now expects an array of values.' );

			/**
			 * Child item thumbnail size.
			 * 
			 * @since 2.6.0
			 * @deprecated 2.7.0
			 * 
			 * @param string $size
			 * @param  obj WC_MNM_Child_Item $child_item of child item
			 * @param  obj WC_Product_Mix_and_Match $product
			 */
			$image_size    = (string) apply_filters( 'wc_mnm_api_child_item_thumbnail_size', $image_size, $child_item, $container_product );
		}

		/**
		 * Child item image sizes.
		 * 
		 * Switching to supporting an array of images here.
		 * 
		 * @since 2.7.0
		 * 
		 * @param string[] $sizes Image sizes.
		 * @param  obj WC_MNM_Child_Item $child_item of child item
		 * @param  obj WC_Product_Mix_and_Match $product
		 */
		$image_sizes    = (array) apply_filters( 'wc_mnm_api_child_item_thumbnail_sizes', array( $image_size ), $child_item, $container_product );

		$images         = array();
		$attachment_ids = array();

		// Add featured image.
		if ( $product->get_image_id() ) {
			$attachment_ids[] = $product->get_image_id();
		}

		// Build image data.
		foreach ( $attachment_ids as $position => $attachment_id ) {
			$attachment_post = get_post( $attachment_id );
			if ( is_null( $attachment_post ) ) {
				continue;
			}

			foreach( $image_sizes as $image_size ) {
				$attachment = wp_get_attachment_image_src( $attachment_id, $image_size );
				if ( ! is_array( $attachment ) ) {
					continue;
				}

				$images[] = array(
					'id'                => (int) $attachment_id,
					'size'				=> $image_size,
					'src'               => current( $attachment ),
					'name'              => get_the_title( $attachment_id ),
					'alt'               => get_post_meta( $attachment_id, '_wp_attachment_image_alt', true ),
				);
			}
		}

		// Set a placeholder image if the product has no images set.
		if ( empty( $images ) ) {
			$images[] = array(
				'id'                => 0,
				'src'               => wc_placeholder_img_src(),
				'name'              => __( 'Placeholder', 'woocommerce-mix-and-match-products' ),
				'alt'               => __( 'Placeholder', 'woocommerce-mix-and-match-products' ),
			);
		}

		return $images;
	}


	/*
	|--------------------------------------------------------------------------
	| Cart Item Functions.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Register parent/child product data into cart/items endpoint.
	 *
	 * @param array  $cart_item
	 * @return array $item_data
	 */
	public static function extend_cart_item_data( $cart_item ) {

		$item_data = array();

		if ( wc_mnm_is_container_cart_item( $cart_item ) ) {

			if ( ! wc_mnm_is_product_container_type( $cart_item['data'] ) ) {
				return $item_data;
			}

			$container = $cart_item['data'];

			// Reset last item key.
			self::$last_child_item_key = false;

			$child_cart_keys = wc_mnm_get_child_cart_items( $cart_item, false, true );

			// Find and store last item.
			if ( $child_cart_keys ) {
				self::$last_child_item_key = end( $child_cart_keys );
			}

			$item_data['child_items']    = $cart_item['mnm_contents'];
			$item_data['container_data'] = array(
				'configuration'         => $cart_item['mnm_config'],
				'is_priced_per_product' => $container->is_priced_per_product(),
				'is_editable'           => apply_filters( 'wc_mnm_show_edit_it_cart_link', true, $cart_item, $cart_item['key'] ),
			);

		} elseif ( wc_mnm_maybe_is_child_cart_item( $cart_item ) ) {

			$container_item = wc_mnm_get_cart_item_container( $cart_item );

			if ( $container_item ) {

				$container = $container_item['data'];

				if ( ! wc_mnm_is_product_container_type( $container ) ) {
					return $item_data;
				}

				// The child item ID is not set in order again, so let's fetch the item by it's product ID.
				$child_item = $container->get_child_item_by_product_id( $cart_item['variation_id'] ? $cart_item['variation_id'] : $cart_item['product_id'] );

				if ( ! $child_item ) {
					return $item_data;
				}

				$item_data['container'] = $cart_item['mnm_container'];

				$child_config_qty = $cart_item['quantity'] / $container_item['quantity'];

				$item_data['child_item_data'] = array(
					'container_id'          => $container_item['product_id'],
					'child_item_id'         => $child_item->get_child_item_id(),
					'child_qty'             => $child_config_qty,
					'is_priced_per_product' => $container->is_priced_per_product(),
					'is_last'               => self::$last_child_item_key === $cart_item['key'],
				);
			}
		}

		return $item_data;
	}

	/**
	 * Register mix and match product schema into cart/items endpoint.
	 *
	 * @return array Registered schema.
	 */
	public static function extend_cart_item_schema() {
		return array(
			'mnm_container'   => array(
				'description' => esc_html__( 'Cart item key of mix and match product that contains this item.', 'woocommerce-mix-and-match-products' ),
				'type'        => array( 'string', 'null' ),
				'context'     => array( 'view', 'edit' ),
				'readonly'    => true,
			),
			'child_items'     => array(
				'description' => esc_html__( 'List of cart item keys grouped by this mix and match product.', 'woocommerce-mix-and-match-products' ),
				'type'        => array( 'array', 'null' ),
				'context'     => array( 'view', 'edit' ),
				'readonly'    => true,
			),
			'container_data'  => array(
				'description' => esc_html__( 'Mix and Match product data.', 'woocommerce-mix-and-match-products' ),
				'type'        => array( 'object', 'null' ),
				'context'     => array( 'view', 'edit' ),
				'readonly'    => true,
			),
			'child_item_data' => array(
				'description' => esc_html__( 'ID of this child item.', 'woocommerce-mix-and-match-products' ),
				'type'        => array( 'object', 'null' ),
				'context'     => array( 'view', 'edit' ),
				'readonly'    => true,
			),
		);
	}

	/**
	 * Aggregates container item prices.
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_container_cart_item_prices( &$item_data, $cart_item ) {

		if ( ! wc_mnm_is_product_container_type( $cart_item['data'] ) || ! $cart_item['data']->is_priced_per_product() ) {
			return;
		}

		$item_data['prices']->raw_prices['price']         = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_price_amount( $cart_item, 'price' ), wc_get_rounding_precision(), PHP_ROUND_HALF_UP );
		$item_data['prices']->raw_prices['regular_price'] = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_price_amount( $cart_item, 'regular_price' ), wc_get_rounding_precision(), PHP_ROUND_HALF_UP );
		$item_data['prices']->raw_prices['sale_price']    = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_price_amount( $cart_item, 'sale_price' ), wc_get_rounding_precision(), PHP_ROUND_HALF_UP );
	}

	/**
	 * Aggregates container item subtotals.
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_container_cart_item_totals( &$item_data, $cart_item ) {

		if ( ! wc_mnm_is_product_container_type( $cart_item['data'] ) || ! $cart_item['data']->is_priced_per_product() ) {
			return;
		}

		$decimals = isset( $item_data['totals']->currency_minor_unit ) ? $item_data['totals']->currency_minor_unit : wc_get_price_decimals();

		$item_data['totals']->line_total        = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_subtotal_amount( $cart_item, 'total' ), $decimals );
		$item_data['totals']->line_total_tax    = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_subtotal_amount( $cart_item, 'tax' ), $decimals );
		$item_data['totals']->line_subtotal     = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_subtotal_amount( $cart_item, 'subtotal' ), $decimals );
		$item_data['totals']->line_subtotal_tax = self::prepare_money_response( WC_Mix_and_Match()->display->get_container_cart_item_subtotal_amount( $cart_item, 'subtotal_tax' ), $decimals );
	}

	/**
	 * Adjust container item quantity limits to prevent child items from being edited.
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_container_cart_item_quantity_limits( &$item_data, $cart_item ) {

		if ( ! wc_mnm_is_product_container_type( $cart_item['data'] ) ) {
			return;
		}

		$child_cart_items = wc_mnm_get_child_cart_items( $cart_item );

		if ( $child_cart_items ) {

			foreach ( $child_cart_items as $child_cart_item_key => $child_cart_item ) {

				// Let's cache this now, as we'll need it later.
				WC_MNM_Helpers::cache_set(
					'child_item_quantity_limits_' . $child_cart_item_key,
					array(
						'multiple_of' => $child_cart_item['quantity'],
						'minimum'     => $child_cart_item['quantity'],
						'maximum'     => $child_cart_item['quantity'],
					)
				);

			}
		}
	}

	/**
	 * Filter container cart item permalink to support cart editing.
	 *
	 * @since 2.0.7
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_container_cart_item_short_description( &$item_data, $cart_item ) {

		if ( wc_mnm_is_container_cart_item( $cart_item ) ) {

			$container = $cart_item['data'];

			if ( ! wc_mnm_is_product_container_type( $container ) ) {
				return;
			}

			if ( apply_filters( 'wc_mnm_show_edit_it_cart_link', true, $cart_item, $cart_item['key'] ) ) {

				$trimmed_short_description = '';

				if ( $item_data['short_description'] ) {
					$trimmed_short_description = '<p class="wc-block-components-product-metadata__description-text">' . wp_trim_words( $item_data['short_description'], 12 ) . '</p>';
				}

				$edit_in_cart_link              = esc_url( $container->get_cart_edit_link( $cart_item ) );
				$item_data['short_description'] = '<p class="wc-block-cart-item__edit"><a class="components-button wc-block-components-button wp-element-button outlined wc-block-cart-item__edit-link contained" href="' . $edit_in_cart_link . '"><span class="wc-block-components-button__text">' . _x( 'Edit selections', 'edit in cart link text', 'woocommerce-mix-and-match-products' ) . '</span></a></p>' . $trimmed_short_description;
			}
		}
	}

	/**
	 * Disable editing for child item quantities.
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_child_cart_item_quantity_limits( &$item_data, $cart_item ) {

		$child_cart_item_key        = $cart_item['key'];
		$child_item_quantity_limits = WC_MNM_Helpers::cache_get( 'child_item_quantity_limits_' . $child_cart_item_key );

		if ( is_null( $child_item_quantity_limits ) ) {
			return;
		}

		$step = $child_item_quantity_limits['multiple_of'];
		$min  = $child_item_quantity_limits['minimum'];
		$max  = $child_item_quantity_limits['maximum'];

		$item_data['quantity_limits']->multiple_of = $step;
		$item_data['quantity_limits']->minimum     = $min;
		$item_data['quantity_limits']->maximum     = $max;
	}

	/**
	 * Disable permalinks for child items.
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_child_cart_item_permalink( &$item_data, $cart_item ) {

		$container_item = wc_mnm_get_cart_item_container( $cart_item );

		if ( $container_item && wc_mnm_is_product_container_type( $container_item['data'] ) ) {
			$item_data['catalog_visibility'] = 'hidden';
		}
	}


	/**
	 * Convert monetary values from WooCommerce to string based integers, using
	 * the smallest unit of a currency.
	 *
	 * @param string|float  $amount
	 * @param int           $decimals
	 * @param int           $rounding_mode
	 * @return string
	 */
	private static function prepare_money_response( $amount, $decimals = 2, $rounding_mode = PHP_ROUND_HALF_UP ) {
		return woocommerce_store_api_get_formatter( 'money' )->format(
			$amount,
			array(
				'decimals'      => $decimals,
				'rounding_mode' => $rounding_mode,
			)
		);
	}

	/*
	|--------------------------------------------------------------------------
	| Callbacks.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Filter store API responses to:
	 *
	 * - aggregate container prices/subtotals;
	 * - keep min/max/step quantity fields in sync.
	 *
	 * @param  $response  WP_REST_Response
	 * @param  $server    WP_REST_Server
	 * @param  $request   WP_REST_Request
	 * @return WP_REST_Response
	 */
	public static function filter_cart_item_data( $response, $server, $request ) {

		if ( is_wp_error( $response ) ) {
			return $response;
		}

		if ( strpos( $request->get_route(), 'wc/store' ) === false ) {
			return $response;
		}

		$data = $response->get_data();

		if ( empty( $data['items'] ) ) {
			return $response;
		}

		if ( ! WC()->cart ) {
			return $response;
		}

		$cart = WC()->cart->get_cart();

		foreach ( $data['items'] as &$item_data ) {

			$cart_item_key = $item_data['key'];
			$cart_item     = isset( $cart[ $cart_item_key ] ) ? $cart[ $cart_item_key ] : null;

			if ( is_null( $cart_item ) ) {
				continue;
			}

			/**
			 * StoreAPI returns the following fields as
			 * - object (/wc/store/v1/cart)
			 * - array (/wc/store/v1/cart/extensions)
			 *
			 * Casting them to objects, to avoid PHP8+ fatal errors.
			 *
			 * @see https://github.com/woocommerce/woocommerce-product-bundles/issues/1096
			 * @see https://github.com/woocommerce/woocommerce-blocks/issues/7275
			 */

			$item_data['quantity_limits'] = (object) $item_data['quantity_limits'];
			$item_data['prices']          = (object) $item_data['prices'];
			$item_data['totals']          = (object) $item_data['totals'];
			$item_data['extensions']      = (object) $item_data['extensions'];

			if ( wc_mnm_is_container_cart_item( $cart_item ) ) {

				self::filter_container_cart_item_prices( $item_data, $cart_item );
				self::filter_container_cart_item_totals( $item_data, $cart_item );
				self::filter_container_cart_item_quantity_limits( $item_data, $cart_item );
				self::filter_container_cart_item_short_description( $item_data, $cart_item );

			} elseif ( wc_mnm_is_child_cart_item( $cart_item ) ) {

				self::filter_child_cart_item_quantity_limits( $item_data, $cart_item );
				self::filter_child_cart_item_permalink( $item_data, $cart_item );
			}
		}

		$response->set_data( $data );

		return $response;
	}

	/**
	 * Whitelist and format container config in Store API context.
	 * 
	 * Need to get the formatted posted config here so that the order item meta will be the same as traditional order. Unfortunately, this does mean we have to get wc_get_product() before Woo initializes it.
	 * Alternatively, we could re-validate the mnm_config on set_mnm_container_cart_item() but currently wp_list_pluck() throws an undefined array key. https://core.trac.wordpress.org/ticket/59774
	 * 
	 * @since 2.6.0
	 *
	 * @throws RouteException
	 *
	 * @param  array  $add_to_cart_data
	 * @param array   $request Add to cart request params including id, quantity, and variation attributes.
	 */
	public static function add_to_cart_data( $add_to_cart_data, \WP_REST_Request $request ) {

		if ( isset( $request['id'] ) && isset( $request['mnm_config'] ) ) {
			$add_to_cart_data['cart_item_data']['mnm_config'] = WC_Mix_and_Match()->cart->get_posted_container_configuration( $request['id'], $request['mnm_config']  );
		}

		return $add_to_cart_data;
	}

	/**
	 * Validate container in Store API context.
	 *
	 * @throws RouteException
	 *
	 * @param  WC_Product  $product
	 * @param array       $request Add to cart request params including id, quantity, and variation attributes.
	 */
	public static function validate_add_to_cart_item( $product, $request ) {

		if ( wc_mnm_is_product_container_type( $product ) ) {
			try {

				$args = array(
					'context'         => 'add-to-cart',
					'throw_exception' => true,
				);

				$quantity = isset( $request['quantity'] ) ? intval( $request['quantity'] ) : 1;
				$config   = isset( $request['cart_item_data'] ) && isset( $request['cart_item_data']['mnm_config'] ) ? $request['cart_item_data']['mnm_config'] : array();

				WC_Mix_and_Match()->cart->validate_container_configuration( $product, $quantity, $config, $args );

			} catch ( Exception $e ) {
				$notice = $e->getMessage();
				throw new RouteException( 'woocommerce_store_api_invalid_container_configuration', wp_kses_post( $notice ) );
			}
		}
	}

	/**
	 * Validate container in Store API context.
	 *
	 * @throws RouteException
	 *
	 * @param  WC_Product  $product
	 * @param  array       $cart_item
	 * @throws RouteException On error.
	 * @throws InvalidCartException Exception if invalid data is detected in the cart.
	 */
	public static function validate_cart_item( $product, $cart_item ) {

		if ( wc_mnm_is_container_cart_item( $cart_item ) ) {
			try {

				WC_Mix_and_Match()->cart->validate_container_in_cart( $cart_item );

			} catch ( Exception $e ) {

				$notice = sprintf( esc_html_x( '&quot;%1$s&quot; cannot be purchased as configured. %2$s', '[Frontend]', 'woocommerce-mix-and-match-products' ), $product->get_name(), $e->getMessage() );

				$cart_errors = new WP_Error( 'woocommerce_store_api_invalid_container_configuration', $notice );

				throw new InvalidCartException(
					'woocommerce_cart_error',
					$cart_errors,
					409
				);

				throw new RouteException( 'woocommerce_store_api_invalid_container_configuration', wp_kses_post( $notice ) );
			}
		}
	}

	/**
	 * Remove quantity inputs from child items in Store API context.
	 *
	 * @param bool $qty_is_editable
	 * @param  WC_Product  $product
	 * @param  array       $cart_item
	 * @return false
	 */
	public static function product_quantity_editable( $qty_is_editable, $product, $cart_item ) {
		if ( wc_mnm_is_child_cart_item( $cart_item ) ) {
			$qty_is_editable = false;
		}
		return $qty_is_editable;
	}

	/**
	 * Prevents access to the checkout block if a container in the cart is misconfigured.
	 * 
	 * Hooked originally to `woocommerce_store_api_checkout_update_order_meta` which @mikejolley suggests was not intended for this purpose does not recommend using in this manner.
	 *
	 * @throws RouteException
	 *
	 * @param  WC_Order  $order
	 * @throws RouteException On error.
	 */
	public static function validate_draft_order( $order ) {

		wc_deprecated_function( 'WC_MNM_Store_API::validate_draft_order', '2.7.4', 'WC_MNM_Store_API::store_api_cart_errors' );

		foreach ( WC()->cart->cart_contents as $cart_item ) {
			self::validate_cart_item( $cart_item[ 'data' ], $cart_item );
		}
		return;
	}

	/**
	 * Prevent removal of child items.
	 *
	 * @throws RouteException
	 *
	 * @param  string   $cart_item_key
	 * @param  WC_Cart  $cart
	 * @return void
	 */
	public static function prevent_child_item_removal( $cart_item_key, $cart ) {

		if ( ! WC_MNM_Core_Compatibility::is_store_api_request( 'cart/remove-item' ) || ! $cart->find_product_in_cart( $cart_item_key ) || ! wc_mnm_is_child_cart_item( $cart->cart_contents[ $cart_item_key ] ) ) {
			return;
		}

		$notice = esc_html( 'This product is part of a mix and match container and cannot be removed independently.', 'woocommerce-mix-and-match-products' );
		throw new RouteException( 'woocommerce_store_api_mnm_child_item', wp_kses_post( $notice ) );
	}

	/**
	 * Filter container cart item permalink to support cart editing.
	 *
	 * @deprecated 2.0.7
	 *
	 * @param array  $item_data
	 * @param array  $cart_item
	 */
	private static function filter_container_cart_item_permalink( &$item_data, $cart_item ) {
		wc_deprecated_function( 'WC_MNM_Store_API::filter_container_cart_item_permalink()', '2.0.7', 'WC_MNM_Store_API::filter_container_cart_item_short_description()' );
		return self::filter_container_cart_item_short_description( $item_data, $cart_item );
	}
}
