<?php
/**
 * The Main WC_Mix_and_Match class.
 *
 * @class    WC_Mix_and_Match
 * @package  WooCommerce Mix and Match
 * @since    1.0.0
 * @version  2.7.9
 */

if ( ! defined( 'ABSPATH' ) ) {
	exit; // Exit if accessed directly.
}

/**
 * Main WC_Mix_and_Match class.
 *
 * The main instance of the plugin.
 *
 * @since  1.0.0
 */
class WC_Mix_and_Match {

	use WC_MNM_Singleton;

	/**
	 * Plugin Version.
	 *
	 * @var str
	 */
	public $version = '2.7.10';

	/**
	 * Required Version of WooCommerce.
	 *
	 * @var str
	 */
	public $required_woo = WC_MNM_REQUIRED_WOO;

	/**
	 * Auto-load in-accessible properties.
	 *
	 * @param  mixed  $key
	 * @return mixed
	 */
	public function __get( $key ) {
		if ( in_array( $key, array( 'compatibility', 'cart', 'order', 'display', 'product_data' ) ) ) {
			$classname = 'WC_MNM_' . ucfirst( $key );
			return call_user_func( array( $classname, 'get_instance' ) );
		}
	}

	/*
	|--------------------------------------------------------------------------
	| Helper Functions.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Get the plugin url.
	 *
	 * @return string
	 */
	public function plugin_url() {
		return untrailingslashit( plugins_url( '/', WC_MNM_PLUGIN_FILE ) );
	}

	/**
	 * Get the plugin path.
	 *
	 * @return string
	 */
	public function plugin_path() {
		return untrailingslashit( plugin_dir_path( WC_MNM_PLUGIN_FILE ) );
	}

	/**
	 * Get the plugin base path name.
	 *
	 * @since  1.2.0
	 *
	 * @return string
	 */
	public function plugin_basename() {
		return plugin_basename( WC_MNM_PLUGIN_FILE );
	}

	/**
	 * Get the file modified time as a cache buster if we're in dev mode.
	 *
	 * @since 2.0.0
	 *
	 * @param  string  $file
	 * @param  string  $version - A version number, handy for mini-extensions to make use of this method.
	 * @return string
	 */
	public function get_file_version( $file, $version = '' ) {
		if ( defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG && file_exists( $file ) ) {
			return filemtime( $file );
		}
		return $version ? $version : $this->version;
	}

	/*
	|--------------------------------------------------------------------------
	| Load Files.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Fire in the hole!
	 */
	public function initialize_plugin() {

		// Load translation files.
		add_action( 'init', array( $this, 'load_plugin_textdomain' ) );

		$this->define_constants();
		$this->includes();

		// Class containing compatibility functions and filters. Needs to be loaded early.
		WC_MNM_Compatibility::get_instance();

		// Admin & AJAX functions and hooks.
		if ( is_admin() ) {
			WC_MNM_Admin::init();
		}

		// Install and upgrade functions.
		WC_MNM_Install::init();

		// Cart-related functions and hooks.
		WC_MNM_Cart::get_instance();

		// Customizer functions and hooks.
		if ( apply_filters( 'wc_mnm_supports_customizer', true ) ) {
			WC_MNM_Customizer::get_instance();
		}

		// Display class.
		WC_MNM_Display::get_instance();

		// Include order-related functions.
		WC_MNM_Order::get_instance();

		// Rest API endpoints and functions.
		WC_MNM_REST_API::init();

		// Product price filters and price-related functions.
		WC_MNM_Product_Prices::init();

		// Order-again related functions.
		WC_MNM_Order_Again::init();

		// Ajax functions and hooks.
		WC_MNM_Ajax::init();

		// Data class.
		WC_MNM_Data::init();

		// Data sync.
		WC_MNM_DB_Sync::init();

		// Receipt Rendering.
		WC_MNM_Receipt_Rendering::init();

		// Include theme-level hooks and actions files. (re-test with Variable MNM).
		add_action( 'after_setup_theme', array( $this, 'theme_includes' ), 11 );

		/**
		 * WooCommerce Mix and Match is fully loaded.
		 */
		do_action( 'wc_mnm_loaded' );
	}

	/**
	 * Constants.
	 *
	 * @since 1.10.0
	 */
	public function define_constants() {
		wc_maybe_define_constant( 'WC_MNM_ABSPATH', trailingslashit( plugin_dir_path( WC_MNM_PLUGIN_FILE ) ) );
		wc_maybe_define_constant( 'WC_MNM_VERSION', $this->version );
		wc_maybe_define_constant( 'WC_MNM_SUPPORT_URL', $this->get_resource_url( 'ticket-form' ) );
	}

	/**
	 * Include required core files used in admin and on the frontend.
	 */
	public function includes() {

		// Class autoloader.
		include_once WC_MNM_ABSPATH . 'includes/class-wc-mnm-autoloader.php';

		// Functions.
		require_once 'wc-mnm-core-functions.php';

		// Product class.
		require_once 'class-wc-product-mix-and-match-legacy.php';
		require_once 'class-wc-product-mix-and-match.php';

	}

	/**
	 * Admin & AJAX functions and hooks.
	 *
	 * @since 1.2.0
	 * @deprecated 2.7.0
	 */
	public function admin_includes() {
		// Handled via autoload now.
	}

	/**
	 * Displays a warning message if version check fails.
	 *
	 * @return string
	 */
	public function admin_notice() {
		wc_deprecated_function( 'WC_Mix_and_Match::admin_notice()', '1.6.0', 'Function is no longer used.' );
		// translators: %1$s is opening link to WordPress plugin site. 2$s is closing link. %3$s is minimum version of WooCommerce.
		echo '<div class="error"><p>' . sprintf( wp_kses_post( __( '<strong>WooCommerce Mix and Match is inactive.</strong> The %1$sWooCommerce plugin%2$s must be active and at least version %3$s for Mix and Match to function. Please upgrade or activate WooCommerce.', 'woocommerce-mix-and-match-products' ), '<a href="http://wordpress.org/extend/plugins/woocommerce/">', '</a>', $this->required_woo ) ) . '</p></div>';
	}

	/**
	 * Include template functions and hooks.
	 */
	public function theme_includes() {
		require_once 'wc-mnm-template-functions.php';
		require_once 'wc-mnm-template-hooks.php';
	}

	/*
	|--------------------------------------------------------------------------
	| Localization.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Make the plugin translation ready.
	 *
	 * Translations should be added in the WordPress language directory:
	 *      - WP_LANG_DIR/plugins/woocommerce-mix-and-match-products-LOCALE.mo
	 *
	 * @since  1.0.0
	 */
	public function load_plugin_textdomain() {
		load_plugin_textdomain( 'woocommerce-mix-and-match-products', false, dirname( plugin_basename( __FILE__ ) ) . '/languages/' );
	}

	/*
	|--------------------------------------------------------------------------
	| Resources.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Returns URL to a doc or support resource.
	 *
	 * @since  2.0.0
	 *
	 * @param  string  $handle
	 * @return string
	 */
	public function get_resource_url( $handle ) {

		switch ( $handle ) {
			case 'ticket-form':
				$resource = 'https://woocommerce.com/my-account/create-a-ticket/?select=853021';
				break;
			case 'unsupported-types':
				$resource = 'https://woocommerce.com/document/woocommerce-mix-and-match-products/config/#h-supported-product-types';
				break;
			case 'docs':
				$resource = 'https://woocommerce.com/document/woocommerce-mix-and-match-products/';
				break;
			case 'updating':
				$resource = 'https://woocommerce.com/document/how-to-update-woocommerce/';
				break;
			case 'outdated-templates':
				$resource = 'https://woocommerce.com/document/fix-outdated-templates-woocommerce/';
				break;
			case 'new-1.3':
				$resource = 'https://woocommerce.com/document/woocommerce-mix-and-match-products/version-1-3';
				break;
			case 'new-2.0':
				$resource = 'https://woocommerce.com/document/woocommerce-mix-and-match-products/version-2-0';
				break;
			default:
				$resource = false;
		}

		return $resource;
	}

	/*
	|--------------------------------------------------------------------------
	| Deprecated.
	|--------------------------------------------------------------------------
	*/

	/**
	 * Main WC_Mix_and_Match instance.
	 *
	 * Ensures only one instance of WC_Mix_and_Match is loaded or can be loaded.
	 *
	 * @return WC_Mix_and_Match Single instance.
	 */
	public static function instance() {
		return self::get_instance();
	}

} // End class: do not remove or there will be no more guacamole for you.
