<?php
/**
 * WooCommerce Mix and Match Autoloader.
 *
 * @package WooCommerce Mix and Match\Classes
 * @since 2.5.0
 * @version 2.7.4
 */

defined( 'ABSPATH' ) || exit;

/**
 * Autoloader class.
 */
class WC_MNM_Autoloader {

	/**
	 * Path to the includes directory.
	 *
	 * @var string
	 */
	private $include_path = '';

	/**
	 * The Constructor.
	 */
	public function __construct() {
		if ( function_exists( '__autoload' ) ) {
			spl_autoload_register( '__autoload' );
		}

		spl_autoload_register( array( $this, 'autoload' ) );

		$this->include_path = untrailingslashit( plugin_dir_path( WC_MNM_PLUGIN_FILE ) ) . '/includes/';
	}

	/**
	 * Take a class name and turn it into a file name.
	 *
	 * @param  string $class Class name.
	 * @return string
	 */
	private function get_file_name_from_class( $class ) {
		$class = str_replace( '_', '-', $class );
		$class = str_replace( 'wc_mix_and_match_', 'wc_mnm_', $class ); // Catch any references to old WC_Mix_and_Match_Something classes.
		return 'class-' . $class . '.php';
	}

	/**
	 * Include a class file.
	 *
	 * @param  string $path File path.
	 * @return bool Successful or not.
	 */
	private function load_file( $path ) {
		if ( $path && is_readable( $path ) ) {
			include_once $path; // nosemgrep: audit.php.lang.security.file.inclusion-arg
			return true;
		}
		return false;
	}

	/**
	 * Auto-load WC classes on demand to reduce memory consumption.
	 *
	 * @param string $class Class name.
	 */
	public function autoload( $class ) {
		$class = strtolower( $class );

		if ( 0 !== strpos( $class, 'wc_mnm_' ) && 0 !== strpos( $class, 'wc_mix_and_match_' ) ) {
			return;
		}

		$file = $this->get_file_name_from_class( $class );
		$path = '';

		switch ( true ) {
			case false !== strpos( $class, 'admin' ):
				$path = $this->include_path . 'admin/';
				break;
	
			case false !== strpos( $class, 'api' ):
				$path = $this->include_path . 'api/';
				break;
	
			case false !== strpos( $class, 'mnm_data' ) || false !== strpos( $class, 'db_sync' ):
				$path = $this->include_path . 'data/';
				break;

			case false !== strpos( $class, 'notes_' ):
				$path = $this->include_path . 'admin/notes/';
				break;
	
			case false !== strpos( $class, 'compatibility' ):
				$path = $this->include_path . 'compatibility/';
				break;
	
			case false !== strpos( $class, 'customizer' ):
				$path = $this->include_path . 'customizer/';
				break;

			case false !== strpos( $class, 'receipt' ):
				$path = $this->include_path . 'receipts/';
				break;

		}

		if ( empty( $path ) || ! $this->load_file( $path . $file ) ) {
			$this->load_file( $this->include_path . $file );
		}
	}
}

new WC_MNM_Autoloader();
