<?php
/**
 * Product data parallel structure for storing product properties.
 *
 * @package  WooCommerce Mix and Match Products/Product Data
 * @since    2.7.0
 * @version  2.7.0
 */

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * WC_MNM_Product_Data class
 */
class WC_MNM_Product_Data {

	use WC_MNM_Singleton;

	/**
	 * @var array - for storing object properties.
	 */
	protected $data = array();

	/**
	 * @var bool - is PHP8 (and WeakMap available).
	 */
	protected $use_weakmap = false;

	/**
	 * Overriding the constructor with a private one prevents calling it directly.
	 */
	private function __construct() {
		$this->use_weakmap = defined( 'PHP_VERSION_ID' ) && PHP_VERSION_ID >= 80000;

		if ( $this->use_weakmap ) {
			$this->data = new WeakMap(); // phpcs:ignore PHPCompatibility.Classes.NewClasses
		}
	}

	/**
	 * Gets product data.
	 *
	 * @param WC_Product   $product
	 * @param string       $key
	 * @param null|string  $default
	 *
	 * @return string
	 */
	public function get( $product, $key, $default = null ) {

		if ( ! is_object( $product ) || ! method_exists( $product, 'get_id' ) ) {
			return $default;
		}

		$identifier = $this->use_weakmap ? $product : spl_object_hash( $product ) . $product->get_id();

		if ( isset( $this->data[ $identifier ] ) && isset( $this->data[ $identifier ][ $key ] ) ) {
			$value = $this->data[ $identifier ][ $key ];
        } else {
			$value = $default;
		}

		return $value;

	}

	/**
	 * Sets product data.
	 *
	 * @param WC_Product  $product
	 * @param string      $key
	 * @param string      $value
	 *
	 */
	public function set( $product, $key, $value ) {

		if ( ! is_object( $product ) || ! method_exists( $product, 'get_id' ) ) {
			return;
		}

		$identifier = $this->use_weakmap ? $product : spl_object_hash( $product ) . $product->get_id();

		if ( ! isset( $this->data[ $identifier ] ) ) {
			$this->data[ $identifier ] = array();
		}

		$this->data[ $identifier ][ $key ] = $value;
	


	}

} // End class.
