<?php
/**
 * Force Sells Compatibility
 *
 * @package  WooCommerce Mix and Match Products/Compatibility
 * @since    2.7.9
 * @version  2.7.10
 */

use KoiLab\WC_Force_Sells\Utilities\Product_Utils;

// Exit if accessed directly.
if ( ! defined( 'ABSPATH' ) ) {
	exit;
}

/**
 * WC_MNM_Force_Sells_Compatibility Class.
 *
 * Adds compatibility with WooCommerce One Page Checkout.
 */
class WC_MNM_Force_Sells_Compatibility {

	public static function init() {

		add_action( 'woocommerce_product_options_related', array( __CLASS__, 'write_panel_tab' ), 20 );
		add_action( 'woocommerce_admin_process_product_object', array( __CLASS__, 'save_product_meta' ) );

		add_action( 'wc_mnm_child_item_details', array( __CLASS__, 'show_force_sell_products_for_child_items' ), 68 );
		add_filter( 'wc_force_sell_product_template_message', array( __CLASS__, 'modify_template_message' ), 10, 2 );
		
		add_filter( 'wc_force_sell_add_to_cart_product', array( __CLASS__, 'maybe_sync_force_sells_on_add_to_cart' ), 10, 2 );
		add_filter( 'wc_force_sell_update_quantity', array( __CLASS__, 'maybe_sync_force_sells_on_update' ), 10, 2 );	
		add_filter( 'wc_force_sell_product_cart_quantity', array( __CLASS__, 'maybe_sync_wc_force_sells_in_store_api' ), 10, 4 );
		add_action( 'woocommerce_add_to_cart', array( __CLASS__, 'add_force_sells_to_cart_for_child_items' ), 20, 6 );

	}

	/*
	|--------------------------------------------------------------------------
	| Admin.
	|--------------------------------------------------------------------------
	*/


	/**
	 * Renders field in Linked Products tab.
	 */
	public static function write_panel_tab() : void {
		global $product_object;

		woocommerce_wp_checkbox(
			array(
				'id'            => '_wc_mnm_sync_to_selected_items',
				'label'         => esc_html__( 'Sync to Mix and Match', 'woocommerce-mix-and-match-products' ),
				'description'   => esc_html__( 'Synced Force Sells are synced quantity to the number of selected child items.', 'woocommerce-mix-and-match-products' ),
				'value'         => wc_bool_to_string( self::is_force_sells_synced_to_child_items( $product_object ) ),
				'wrapper_class' => 'show_if_mix-and-match',
			)
		);

	}

	/**
	 * Save meta when product is saved.
	 *
	 * @param object $product
	 */
	public static function save_product_meta( $product ) {
		// phpcs:disable WordPress.Security.NonceVerification
		if ( isset( $_POST[ '_wc_mnm_sync_to_selected_items' ] ) ) {
			$product->update_meta_data( '_wc_mnm_sync_to_selected_items', 'yes' );
		} else {
			$product->delete_meta_data( '_wc_mnm_sync_to_selected_items' );
		}
		// phpcs:enable WordPress.Security.NonceVerification
	}

	/*
	|--------------------------------------------------------------------------
	| Single product display
	|--------------------------------------------------------------------------
	*/

	/**
	 * Display Mix and Match child product force sells
	 *
	 * @param obj WC_MNM_Child_Item $child_item of child item
	 */
	public static function show_force_sell_products_for_child_items( $child_item ) {

		$child_product = $child_item->get_product();

		if ( ! $child_product ) {
			return;
		}

		$force_sells = \KoiLab\WC_Force_Sells\Utilities\Product_Utils::get_valid_force_sells( (int) $child_product->get_id(), '', 'objects' );

		if ( ! $force_sells ) {
			return;
		}

		WC_Force_Sells::instance()->show_force_sell_products( $child_product );

	}

	/**
	 * Maybe change the force sells template message
	 *
	 * @param string $message Template message.
	 * @param WC_Product  $product  The produce object.
	 * @return string
	 */
	public static function modify_template_message( $message, $product ) {
		if ( self::is_force_sells_synced_to_child_items( $product ) ) {
			$message = '<p>' . esc_html__( 'This will also add the following products to your cart, one for each item selected', 'woocommerce-mix-and-match-products' ) . '</p>';
		}

		return $message;
	}
	
	/*
	|--------------------------------------------------------------------------
	| Cart Functions.
	|--------------------------------------------------------------------------
	*/


	/**
	 * Sync the container force sells to the number of selected child items on Add to Cart
	 *
	 * @param  array   $params {
	 * 	@type int 'id' - The ID of the product being forced to the cart
	 *  @type int 'quantity' - The quantity of the product being forced to the cart.
	 *  @type int 'variation_id' - The variation ID of the product being forced to the cart.
	 *  @type array 'variation' - The variation attribute data of the product being forced to the cart.
	 * }
	 * @return  array    $params
	 */
	public static function maybe_sync_force_sells_on_add_to_cart( $params, $cart_item ) {

		$factor = self::child_items_quantity_factor( $cart_item );
		$params['quantity'] = $cart_item['quantity'] * $factor;
		return $params;
		
	}

	/**
	 * Sync the container force sells to the number of selected child items on cart Update
	 *
	 * @param  int $quantity - The quantity of the product being forced to the cart.
	 * @param array $cart_item - This is the forced product item in the cart
	 * @return int -  Updated quantity
	 */
	public static function maybe_sync_force_sells_on_update( $quantity, $cart_item ) {

		if ( isset( $cart_item['forced_by'] ) ) {

			$forced_by_cart_item = WC()->cart->get_cart_item( $cart_item['forced_by'] );
			$factor = self::child_items_quantity_factor( $forced_by_cart_item );
			$quantity = $forced_by_cart_item['quantity'] * $factor;		

		}

		return $quantity;
	}

	/**
	 * Sync the container force sells to the number of selected child items in the Store API.
	 * 
	 * @see woocommerce-force-sells/src/Cart.php
	 * 
	 * @param float|int $qty the forced quantity
	 * @param WC_Product $product the product object
	 * @param WC_Product $forced_by_product the product that forced the quantity
	 * @param array<string,mixed> $forced_by_cart_item the cart item that forced the quantity
	 * @return float|int
	 */
	public static function maybe_sync_wc_force_sells_in_store_api( $qty, $product, $forced_by_product, $forced_by_cart_item ) {
		$factor = self::child_items_quantity_factor( $forced_by_cart_item );
		$qty = $forced_by_cart_item['quantity'] * $factor;

		return $qty;
	}

	/**
	 * Calculate the number of child items in a given container if syncing force sells to child items.
	 * 
	 * @param array $cart_item the container cart item
	 * @return int The number of child items in the cart related to the given container
	 */
	private static function child_items_quantity_factor( $cart_item ) {

		$factor = 1;

		if ( wc_mnm_is_container_cart_item( $cart_item ) ) {

			if ( self::is_force_sells_synced_to_child_items( $cart_item['data'] ) ) {
				
				// This is a container cart item, so we need to get the child items in the cart.
				$child_cart_items = wc_mnm_get_child_cart_items( $cart_item );

				if ( ! empty( $child_cart_items ) ) {
					$child_qty = 0;
					foreach( $child_cart_items as $child_cart_item ) {
						$child_qty += $child_cart_item['quantity'];
					}

					$factor = $child_qty / $cart_item['quantity'];
				}

			}
		
		}

		return $factor;
	}

	/**
	 * Add force sells for MNM contents to cart after container.
	 *
	 * @param  string   $container_cart_key
	 * @param  int      $product_id
	 * @param  int      $quantity
	 * @param  int      $variation_id
	 * @param  array    $variation
	 * @param  array    $cart_item_data
	 */
	public static function add_force_sells_to_cart_for_child_items( $container_cart_key, $product_id, $quantity, $variation_id, $variation, $cart_item_data ) {

		$container_cart_item = WC()->cart->get_cart_item( $container_cart_key );

		foreach( wc_mnm_get_child_cart_items( $container_cart_item ) as $child_cart_item ) {
			$child_cart_key = $child_cart_item['key'];
			$child_product_id = $child_cart_item['product_id'];
			$child_quantity = $child_cart_item['quantity'];
			\WC_Force_Sells::instance()->add_force_sell_items_to_cart( $child_cart_key, $child_product_id, $child_quantity );
		}	

	}

	/**
	 * Is this Mix and Match product using synced force sells?
	 * 
	 * @param mixed int|WC_Product - The product ID or object.
	 */
	private static function is_force_sells_synced_to_child_items( $product ) : bool {
		if ( is_numeric( $product ) ) {
			$product = wc_get_product( $product );
		}

		if ( ! $product || ! $product->is_type( 'mix-and-match' ) ) {
			return false;
		}

		return wc_string_to_bool( $product->get_meta( '_wc_mnm_sync_to_selected_items', true ) );
	}

}

WC_MNM_Force_Sells_Compatibility::init();
